@@ -329,7 +329,8 @@ class BaseExperiment:
         if self.model is None:
             raise AssertionError("Model must be defined first. Use .model(model_type, **hparams) to declare a model.")
         
-        del self.data_module
+        if self.data_module:
+            del self.data_module
         self.data_module = None
         data_module = None
         if self.n_classes is None and stratify:
@@ -488,6 +489,7 @@ class BaseExperiment:
                 self.y_train, self.y_val = y_train, y_val
             else:
                 raise ValueError("X_test and y_test should both be NoneType object or non-empty numpy array.")
+        self.data_module = data_module
         return data_module
     
     
@@ -537,6 +539,7 @@ class BaseExperiment:
         self,
         sample_weight : Optional[Union["np.ndarray", "torch.Tensor"]] = None,
         fit_or_cv : str = "FIT",
+        stratify : bool = True,
         **config        
     )->Union["pl.LightningModule", "BaseEstimator"]:
         r"""Method for training the model.
@@ -546,7 +549,11 @@ class BaseExperiment:
             sample_weight (np.ndarray or torch.Tensor, optional): If specified, this will be the weights for each sample.
                 Default: ``None``
             
+            stratify (bool): Whether to use label-stratification in data splitting. Forced to be `False` when it is a regression task.
+                Default: ``True``
+
             fit_or_cv (str): Flag. "FIT" if in actual training, "CV" if in cross validation.
+                Default: ``"FIT"``
 
             **config (Any): Trainer configurations.
 
@@ -561,7 +568,8 @@ class BaseExperiment:
         
         self.fit_or_cv = fit_or_cv
         self.update_log_path()
-        self.data_module = self.get_data_module()
+        if self.data_module is None:
+            self.data_module = self.get_data_module(stratify=stratify)
         if self.logger:
             self.logger.close()
             del self.logger
