@@ -1,5 +1,7 @@
 import torch
 import torch.nn as nn
+from torch.optim.lr_scheduler import OneCycleLR, StepLR, ExponentialLR, ConstantLR
+from torch.optim.lr_scheduler import LinearLR,  ReduceLROnPlateau
 import pytorch_lightning as pl
 from astrape.utilities.utils import *
 from astrape.exceptions.exceptions import *
@@ -14,6 +16,7 @@ applicable_val_metric_list = APPLICABLE_VAL_METRIC_LIST
 applicable_optimizer_type_list = APPLICABLE_OPTIMIZER_TYPE_LIST
 
 
+    
 class BaseNet(pl.LightningModule):
     r"""BaseNet. Use this for models that don't deal with image data. 
 
@@ -72,9 +75,38 @@ class BaseNet(pl.LightningModule):
         self.sample_weight = sample_weight
 
         config.update({'l1_strength' : l1_strength})
-            
+        if 'scheduler' not in config.keys():
+            config.update({"scheduler" : OneCycleLR})
         self.save_hyperparameters(*config)
         
+    def get_scheduler(self, optimizer, scheduler_class, **kwargs):
+
+        if scheduler_class is None:
+            return ConstantLR(optimizer, factor=1)
+        if scheduler_class == OneCycleLR:
+            if 'max_lr' not in kwargs.keys():
+                kwargs.update({'max_lr' : 0.1})
+            if 'epochs' not in kwargs.keys() and 'total_steps' not in kwargs.keys() and 'steps_per_epoch' not in kwargs.keys():
+                kwargs.update({'total_steps' : self.trainer.max_steps})
+            
+
+        elif scheduler_class == StepLR:
+            if 'step_size' not in kwargs.keys():
+                kwargs.update({'step_size': 30})
+            if 'gamma' not in kwargs.keys():
+                kwargs.update({'gamma' : 0.1}) 
+        elif scheduler_class == ConstantLR:
+            pass
+        elif scheduler_class == LinearLR:
+            pass
+        elif scheduler_class == ExponentialLR:
+            if 'gamma' not in kwargs.keys():
+                kwargs.update({'gamma':0.5})
+        elif scheduler_class == ReduceLROnPlateau:
+            pass
+
+        scheduler = scheduler_class(optimizer=optimizer, **kwargs)
+        return scheduler
 
     @property
     def criterion(self):
@@ -142,15 +174,12 @@ class BaseNet(pl.LightningModule):
             self.log('val/loss', loss)
             return {'loss' : loss}
         else:    
-            loss, acc, auc = self.criterion(Yhat.squeeze().type_as(y_batch), y_batch)
-
+            loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch)
+            self.log('val/loss', loss)
+            self.log('val/acc', acc)
+            self.log('val/auc', auc)
             return {'loss' : loss, 'acc' : acc, 'auc' : auc}
-    def validation_epoch_end(self, outputs):
-        loss = torch.stack([x["loss"] for x in outputs]).mean()
-        acc = torch.stack([x["acc"] for x in outputs]).mean()
-        auc = torch.stack([x["auc"] for x in outputs]).mean()
-        log_dict = {'val/loss' : loss, 'val/acc' : acc, 'val/auc' : auc}
-        return {"log": log_dict, "progress_bar" : log_dict}
+
 
     def predict_step(self, pred_batch, batch_idx, dataloader_idx=0):
         x_batch, y_batch = pred_batch
@@ -225,9 +254,18 @@ class BaseNet(pl.LightningModule):
                                      weight_decay=self.hparams.weight_decay)
         elif self.hparams.optimizer_type == "sgd":
             optimizer = torch.optim.SGD(self.parameters(),
-                                        lr=self.hparams.lr)
-        
-        return optimizer
+                                        lr=self.hparams.lr,
+                                        momentum=self.hparams.momentum,
+                                        weight_decay=self.hparams.weight_decay)
+
+        if 'scheduler' not in self.config.keys():
+            self.config.update({'scheduler' : None})
+
+        scheduler_dict = {
+            "scheduler" : self.get_scheduler(optimizer=optimizer, scheduler_class=self.config["scheduler"]),
+            "interval" : "step"
+        }
+        return {"optimizer": optimizer, "lr_scheduler":scheduler_dict}
 
 
 class ConvNet(pl.LightningModule):
@@ -278,8 +316,37 @@ class ConvNet(pl.LightningModule):
         self.n_classes = n_classes
         self.sample_weight = sample_weight
         self.dataformats = dataformats
+        self.config = config
         self.save_hyperparameters(*config)
-        
+
+    def get_scheduler(self, optimizer, scheduler_class, **kwargs):
+
+        if scheduler_class is None:
+            return ConstantLR(optimizer, factor=1)
+        if scheduler_class == OneCycleLR:
+            if 'max_lr' not in kwargs.keys():
+                kwargs.update({'max_lr' : 0.1})
+            if 'epochs' not in kwargs.keys() and 'total_steps' not in kwargs.keys() and 'steps_per_epoch' not in kwargs.keys():
+                kwargs.update({'total_steps' : self.trainer.max_steps})
+            
+
+        elif scheduler_class == StepLR:
+            if 'step_size' not in kwargs.keys():
+                kwargs.update({'step_size': 30})
+            if 'gamma' not in kwargs.keys():
+                kwargs.update({'gamma' : 0.1}) 
+        elif scheduler_class == ConstantLR:
+            pass
+        elif scheduler_class == LinearLR:
+            pass
+        elif scheduler_class == ExponentialLR:
+            if 'gamma' not in kwargs.keys():
+                kwargs.update({'gamma':0.5})
+        elif scheduler_class == ReduceLROnPlateau:
+            pass
+
+        scheduler = scheduler_class(optimizer=optimizer, **kwargs)
+        return scheduler
 
     @property
     def criterion(self):   
@@ -321,10 +388,6 @@ class ConvNet(pl.LightningModule):
             raise ValueError(f"Wrong dimension {dims} for the input.")
 
         return x_batch
-
-    def on_before_batch_transfer(self, batch: Any, dataloader_idx=0):
-        batch[0] = self.reshape_x(batch[0])
-        return batch
     
     def on_after_batch_transfer(self, batch, dataloader_idx=0):
         # modify this part if you want to perform model reweigthing.
@@ -333,10 +396,10 @@ class ConvNet(pl.LightningModule):
     
     def training_step(self, train_batch, batch_idx):
         x_batch, y_batch = train_batch
-        x_batch = self.reshape_x(x_batch)
+        #x_batch = self.reshape_x(x_batch).type_as(x_batch)
         y_batch = self.sample_weight * y_batch
-
         Yhat = self.forward(x_batch)
+        
         if self.n_classes is None:
             loss = self.criterion(Yhat.squeeze(), y_batch)
             self.log('train/loss', loss)
@@ -349,15 +412,13 @@ class ConvNet(pl.LightningModule):
             self.log('train/loss', loss)
             self.log('train/acc', acc)
             self.log('train/auc', auc)
-            log_dict = {'train/loss' : loss, 'train/acc' : acc, 'train/auc' : auc}
-            return {'loss' : loss, 'acc' : acc, 'auc' : auc, "progress_bar" : log_dict}
+            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
 
-
-        
     def validation_step(self, val_batch, batch_idx):
         x_batch, y_batch = val_batch
-        x_batch = self.reshape_x(x_batch)
+        #x_batch = self.reshape_x(x_batch).type_as(x_batch)
         Yhat = self.forward(x_batch)
+        
         if self.n_classes is None:
             loss = self.criterion(Yhat.squeeze(), y_batch)
             self.log('val/loss', loss)
@@ -367,20 +428,16 @@ class ConvNet(pl.LightningModule):
                 loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))
             else:
                 loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.FloatTensor))
-
+            self.log('val/loss', loss)
+            self.log('val/acc', acc)
+            self.log('val/auc', auc)
             return {'loss' : loss, 'acc' : acc, 'auc' : auc}
 
-    def validation_epoch_end(self, outputs):
-        loss = torch.stack([x["loss"] for x in outputs]).mean()
-        acc = torch.stack([x["acc"] for x in outputs]).mean()
-        auc = torch.stack([x["auc"] for x in outputs]).mean()
-        log_dict = {'val/loss' : loss, 'val/acc' : acc, 'val/auc' : auc}
-        return {"log": log_dict, "progress_bar" : log_dict}
-
     def predict_step(self, pred_batch, batch_idx, dataloader_idx=0):
         x_batch, y_batch = pred_batch
-        x_batch = self.reshape_x(x_batch)
+        #x_batch = self.reshape_x(x_batch).type_as(x_batch)
         Yhat = self.forward(x_batch)
+        
         if self.n_classes is None:
             loss = self.criterion(Yhat.squeeze(), y_batch)
             self.log('pred/loss', loss)
@@ -397,8 +454,9 @@ class ConvNet(pl.LightningModule):
 
     def test_step(self, test_batch, batch_idx):
         x_batch, y_batch = test_batch
-        x_batch = self.reshape_x(x_batch)
+        #x_batch = self.reshape_x(x_batch).type_as(x_batch)
         Yhat = self.forward(x_batch)
+        
         if self.n_classes is None:
             loss = self.criterion(Yhat.squeeze(), y_batch)
             self.log('test/loss', loss)
@@ -429,15 +487,25 @@ class ConvNet(pl.LightningModule):
                 nn.init.constant_(module.bias, 0)
 
     def configure_optimizers(self):
+
         if self.hparams.optimizer_type == "adam":
             optimizer = torch.optim.Adam(self.parameters(),
                                      lr=self.hparams.lr,
                                      weight_decay=self.hparams.weight_decay)
         elif self.hparams.optimizer_type == "sgd":
             optimizer = torch.optim.SGD(self.parameters(),
-                                        lr=self.hparams.lr)
+                                        lr=self.hparams.lr,
+                                        momentum=self.hparams.momentum,
+                                        weight_decay=self.hparams.weight_decay)
 
-        return optimizer
+        if 'scheduler' not in self.config.keys():
+            self.config.update({'scheduler' : None})
+
+        scheduler_dict = {
+            "scheduler" : self.get_scheduler(optimizer=optimizer, scheduler_class=self.config["scheduler"]),
+            "interval" : "step"
+        }
+        return {"optimizer": optimizer, "lr_scheduler":scheduler_dict}
 
 
 class SegNet(pl.LightningModule):
@@ -481,6 +549,7 @@ class SegNet(pl.LightningModule):
         self.n_classes = n_classes
         self.sample_weight = sample_weight
         self.dataformats = dataformats
+        self.config = config
         self.save_hyperparameters(*config)
         
 
@@ -558,11 +627,11 @@ class SegNet(pl.LightningModule):
                 loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.LongTensor))
             else:
                 loss, acc, auc = self.criterion(Yhat.squeeze(), y_batch.type(torch.FloatTensor))
-            self.log({'val/loss' : loss})
-            self.log({'val/acc' : acc})
-            self.log({'val/auc' : auc})
+            self.log('val/loss', loss)
+            self.log('val/acc', acc)
+            self.log('val/auc', auc)
             log_dict = {'val/loss' : loss, 'val/acc' : acc, 'val/auc' : auc}
-            return {"progress_bar" : log_dict}
+            return {'loss' : loss, 'acc' : acc, 'auc' : auc}
 
 
     def predict_step(self, pred_batch, batch_idx, dataloader_idx=0):
@@ -615,17 +684,55 @@ class SegNet(pl.LightningModule):
             nn.init.kaiming_normal_(module.weight, mode="fan_out", nonlinearity="relu")
             if module.bias is not None:
                 nn.init.constant_(module.bias, 0)
+    def get_scheduler(self, optimizer, scheduler_class, **kwargs):
+
+        if scheduler_class is None:
+            return ConstantLR(optimizer, factor=1)
+        if scheduler_class == OneCycleLR:
+            if 'max_lr' not in kwargs.keys():
+                kwargs.update({'max_lr' : 0.1})
+            if 'epochs' not in kwargs.keys() and 'total_steps' not in kwargs.keys() and 'steps_per_epoch' not in kwargs.keys():
+                kwargs.update({'total_steps' : self.trainer.max_steps})
+            
+
+        elif scheduler_class == StepLR:
+            if 'step_size' not in kwargs.keys():
+                kwargs.update({'step_size': 30})
+            if 'gamma' not in kwargs.keys():
+                kwargs.update({'gamma' : 0.1}) 
+        elif scheduler_class == ConstantLR:
+            pass
+        elif scheduler_class == LinearLR:
+            pass
+        elif scheduler_class == ExponentialLR:
+            if 'gamma' not in kwargs.keys():
+                kwargs.update({'gamma':0.5})
+        elif scheduler_class == ReduceLROnPlateau:
+            pass
+
+        scheduler = scheduler_class(optimizer=optimizer, **kwargs)
+        return scheduler
 
     def configure_optimizers(self):
+
         if self.hparams.optimizer_type == "adam":
             optimizer = torch.optim.Adam(self.parameters(),
                                      lr=self.hparams.lr,
                                      weight_decay=self.hparams.weight_decay)
         elif self.hparams.optimizer_type == "sgd":
             optimizer = torch.optim.SGD(self.parameters(),
-                                        lr=self.hparams.lr)
+                                        lr=self.hparams.lr,
+                                        momentum=self.hparams.momentum,
+                                        weight_decay=self.hparams.weight_decay)
 
-        return optimizer
+        if 'scheduler' not in self.config.keys():
+            self.config.update({'scheduler' : None})
+
+        scheduler_dict = {
+            "scheduler" : self.get_scheduler(optimizer=optimizer, scheduler_class=self.config["scheduler"]),
+            "interval" : "step"
+        }
+        return {"optimizer": optimizer, "lr_scheduler":scheduler_dict}
 """
 class GenerativeNet(pl.LightningModule):
     def __init__(
